# Copyright (C) 2017-2018 Aleksandr Popov

# This program is free software: you can redistribute it and/or modify
# it under the terms of the Lesser GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# Lesser GNU General Public License for more details.

# You should have received a copy of the Lesser GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import scipy.fftpack as fftpack
import scipy.signal as sig
import numpy as np

# TODO: use calculated window, not string
# TODO: replace save_energy to some positive argument. Something like save_amplitude or fit_amplitude.
# TODO: Is the default value for save_energy correct?
def spectrum(x, fs=1, window='hamming', one_side=False, return_amplitude=True, extra_len=None, save_energy=False):
    """
    Return the Fourier spectrum of signal.

    Parameters
    ----------
    x : array_like
        Signal values
    fs : float
        Sampling frequency (Hz)
    window : str
        Window.
    one_side : boolean
        If True, the one-side spectrum is calculated (default value is False)
    return_amplitude : boolean
        If True, the amplitude spectrum is calculated
    save_energy : boolean
        If True, the result of FFT has the same energy as signal. 
        If False, the X (spectrum) is multiplied to 2/len(x). Use False if you want to see 
        the correct amplitude of components in spectrum.

    Returns
    -------
    X : np.ndarray of complex numbers
        Spectrum
    f_X : np.ndarray of floats
        Frequency values (Hz)

    """
    # window signal
    win = sig.get_window(window, len(x))
    xx = x * win * len(win)/sum(win)
    # FFT
    if extra_len:
        n = max(extra_len, len(x))
    else:
        n = len(x)
    #X = 2 * fftpack.fft(xx, n) / len(x)
    X = fftpack.fft(xx, n)
    if not save_energy:
        X *= 2/len(x)
    f_X = np.fft.fftfreq(len(X), 1/fs)
    # one-side
    if one_side:
        ind = f_X>=0
        f_X = f_X[ind]
        X = X[ind]
    # calc amplitude
    if return_amplitude: # TODO: result = 'amplitude|phase|None'
        X = abs(X)
    return X, f_X

# TODO: use calculates window, not string
# TODO: default values
def stft(x, fs=1, nseg=256, nstep=None, window='hamming', nfft=None, padded=False):
    """
    Return result of short-time fourier transform.

    Parameters
    ----------
    x : numpy.ndarray
        Signal.
    fs : float 
       Sampling frequency (Hz).
    nseg : int
        Length of segment (in samples).
    nstep : int
        Optional. Length of step (in samples). If not setted then equal to nseg//2.
    window : str
        Window.
    nfft : int 
        Length of the FFT. If None or less than nseg, the FFT length is nseg.

    Returns
    -------

    Xs : numpy.ndarray
        Result of STFT, two-side spectrums.
    
    """
    if not nstep:
        nstep = nseg//2 # TODO: test it
    # TODO: consider nstep in padding
    xx = x.copy()
    if padded:
        L = len(xx) + (nseg - len(xx) % nseg) % nseg
        z = np.zeros(L)
        z[:len(xx)] = xx
        xx = z

    Xs=[]
    for i in range(0, len(xx)-nseg + 1, nstep):
        seg = xx[i : i+nseg]
        X = spectrum(seg, fs, extra_len=nfft, window=window, save_energy=True)[0]
        Xs.append(X)
    return np.array(Xs) # TODO: return times too

# TODO: only here fs has default value
def calc_specgram(x, fs=1, t=[], nseg=256, nstep=None, freq_bounds=None, extra_len=None):
    """
    Return spectrogram data prepared to further plotting.

    Parameters
    ----------
    x : array_like
        Signal values
    fs : float
        Sampling frequency (Hz)
    t : array_like
        Time values (sec)
    nseg : integer
        Length of window (number of samples)
    nstep : integer
        Length of step between Fourier transforms
    freq_bounds : tuple of 2 float
        Bounds of showed band
    extra_len : integer
        Number of values using for fft

    Return
    ------
    Xs : np.ndarray
        Array of spectrums
    t_new : np.ndarray
        Time values

    """
    if len(x) < nseg:
        return [], []
    if len(t)==0:
        t = np.linspace(0, (len(x)-1)*fs, len(x))
    else:
        fs = 1/(t[1] - t[0])
    if not nstep:
        nstep = nseg//2
    # TODO: test with None extra_len
    Xs = 2*stft(x=x, fs=fs, nseg=nseg, nstep=nstep, nfft=extra_len, padded=True)
    if freq_bounds:
        freqs = np.fft.fftfreq(len(Xs[0]), 1/fs)
        ind = (freqs>=freq_bounds[0])&(freqs<=freq_bounds[1]) # TODO: thick about it, maybe mistake here
        Xs = Xs[:,ind]
    t_new = np.linspace(t[nseg-1], t[-1], len(Xs))
    return np.transpose(Xs), t_new # TODO: consider mirroring when plotted



# Copyright (C) 2017-2018 Aleksandr Popov

# This program is free software: you can redistribute it and/or modify
# it under the terms of the Lesser GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# Lesser GNU General Public License for more details.

# You should have received a copy of the Lesser GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

""" This module implements the base class for online filters. """

print("Warning! Deprecated dsplab.online.")

import numpy as np
from collections import deque

# Compatibility
from dsplab.activity import OnlineFilter # TODO: remove in new version

pi = 3.141592653589793
pi2 = 2*pi

def unwrap_point(w):
    """ Unwrap angle (for signle value). """
    if w < -pi:
        return w + pi2
    if w > pi:
        return w - pi2
    return w



# Copyright (C) 2017-2018 Aleksandr Popov

# This program is free software: you can redistribute it and/or modify
# it under the terms of the Lesser GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# Lesser GNU General Public License for more details.

# You should have received a copy of the Lesser GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np
from scipy.signal import get_window

def envelope_by_extremums(x, fs = 1, t = []):
    """ 
    Calculate envelope by local extremums of signals

    Parameters
    ----------
    x : array_like
        Signal values
    fs : float
        Sampling frequency
    t : array_like
        Time values. Use it for unregular discretized input signal

    Returns
    --------
    
    x_new : np.array
        Damping values
    t_new : np.array
        Time values

    """
    if len(t) == 0:
        t = np.linspace(0, (len(x)-1)/fs, len(x))
    t_new = []
    x_new = []
    x = abs(x)
    for x_left, x_central, x_right, t_central in zip(x[:-2], x[1:-1], x[2:], t[1:-1]):
        if (x_left < x_central) and (x_central >= x_right):
            t_new.append(t_central)
            x_new.append(x_central)
    if x[-1] > x[-2]:
        t_new.append(t[-1])
        x_new.append(x[-1])
    return np.array(x_new), np.array(t_new)

def digital_hilbert_filter(ntaps=101, window='hamming'):
    """
    Calculate digital hilbert filter.


    Parameters
    ----------
    ntaps : integer
        Length of filter.
    window : str
        Window. Default is 'hamming'.
    
    Returns
    -------
    h : np.array
        Filter.

    """
    if ntaps%2 == 0:
        raise ValueError("ntaps of digital Hilbert filter must be odd.")
    L = ntaps//2
    h = [1/np.pi/k * (1 - np.cos(np.pi * k)) for k in range(-L, 0)]
    h += [0]
    h += [1/np.pi/k * (1 - np.cos(np.pi * k)) for k in range(1, L+1)]
    w = get_window(window, ntaps)
    h *= w
    return h

def hilbert_filter(L):
    """
    Deprecated in 0.16. Calculate digital hilbert filter smoothed with Hamming's window
 

    Parameters
    ----------
    L : integer
        Length of half of filter
    
    Returns
    -------
    h : np.array
        Filter with length = 2*L + 1

    """
    h = [1/np.pi/k * (1 - np.cos(np.pi * k)) for k in range(-L, 0)]
    h += [0]
    h += [1/np.pi/k * (1 - np.cos(np.pi * k)) for k in range(1, L+1)]
    return np.array(h)*np.hamming(2*L+1)



# Copyright (C) 2017-2018 Aleksandr Popov

# This program is free software: you can redistribute it and/or modify
# it under the terms of the Lesser GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# Lesser GNU General Public License for more details.

# You should have received a copy of the Lesser GNUGeneral Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np
import numpy.linalg as linalg

def prony_decomp(x, L):
    """ 
    Prony decomposition of signal.

    Parameters
    ----------
    x : array_like
        Signal values
    L : integer
        Number of components. 2*L must be less tham length of x
    
    Returns
    -------
    ms : np.array
        Mu-values
    cs : np.array
        C-values
    es : np.array
        Components

    """
    N = len(x)
    if 2*L > N:
        return None
    d = np.array([x[i] for i in range(L, N)])
    D = []
    for i in range(L, N):
        D_row = []
        for j in range(0, L):
            D_row.append(x[i-j-1])
        D.append(np.array(D_row))
    D = np.array(D)
    a = linalg.lstsq(D, d)[0] # TODO: think here if mistake

    p = np.array([1] +[-ai for ai in a])
    ms = np.roots(p)

    d = np.array([x[i] for i in range(0, N)])
    D = []
    for i in range(0, N):
        D_row = []
        for j in range(0, L):
            D_row.append(ms[j]**i)
        D.append(np.array(D_row))
    D = np.array(D)

    cs = linalg.lstsq(D, d)[0] # TODO: think here if mistake

    es = []
    for i in range(0, L):
        e = [cs[i]*(ms[i]**k) for k in range(0, N)]
        es.append(np.array(e).real)
    es = np.array(es)

    return ms, cs, es



import numpy as np

print("Warning! dsplab.io deprecated.")

def read_signal_from_csv(file_name, t_column=0, x_column=1, start_line=1, fs=None, delimiter=';'):
    """ Deprecated. """
    data = np.genfromtxt(file_name, delimiter=delimiter)
    data = data.transpose()
    x = np.array(data[x_column][start_line:])
    if not fs:
        t = np.array(data[t_column][start_line:])
    else:
        t = np.linspace(0, (len(x)-1)/fs, len(x))
    return x, t



# Copyright (C) 2017-2018 Aleksandr Popov

# This program is free software: you can redistribute it and/or modify
# it under the terms of the Lesser GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# Lesser GNU General Public License for more details.

# You should have received a copy of the Lesser GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import scipy.signal as sig
import scipy.fftpack as fftpack
import numpy as np

def _stupid_filter(x, fs, fr):
    """
    Filter signal using setted frequency response.

    Parameters
    ----------
    x : array_like
        Signal values
    fs : float
        Sampling frequency
    fs : np.array
        Frequency response of ideal filter

    Returns
    -------
    xf : np.array
        Filteres signal

    """
    _x = x * sig.tukey(len(x))
    X = fftpack.fft(_x)
    return np.real(fftpack.ifft(X * fr))

def stupid_lowpass_filter(x, fs, cutoff):
    """
    Return low-pass filtered signal.

    Parameters
    ----------
    x : array_like
        Signal values
    fs : float
        Sampling frequency
    cutoff : float
        Cutoff frequency

    Returns
    -------
    xf : np.array
        Filteres signal

    """
    fr = np.zeros(len(x))
    f = np.fft.fftfreq(len(x), 1/fs)
    ind = abs(f)<=cutoff
    fr[ind] = 1
    return  _stupid_filter(x, fs, fr)

def stupid_bandpass_filter(x, fs, bandpass):
    """
    Return low-pass filtered signal.

    Parameters
    ----------
    x : array_like
        Signal values
    fs : float
        Sampling frequency
    bandpass : np.array of 2 floats
        Bounds of bandpass (Hz)

    Returns
    -------
    xf : np.array
        Filteres signal

    """
    fr = np.zeros(len(x))
    f = np.fft.fftfreq(len(x), 1/fs)
    ind = (abs(f) >= bandpass[0])&(abs(f)<=bandpass[1])
    fr[ind] = 1
    return  _stupid_filter(x, fs, fr)

def butter_filter(x, fs, freqs, order, btype='band'):
    """ 
    Butterworth filter

    Parameters
    ----------
    x : array_like
        Signal values
    fs : float
        Sampling frequency (Hz)
    freqs : array_like
        One or two frequencies
    order : integer
        Order of filter
    btype : str (band|lowpass)
        Type of filter

    Returns
    -------
    xf : np.array
        filtered signal

    """
    nyq = 0.5 * fs
    freqs = np.array(freqs)
    freqs /= nyq
    b, a = sig.butter(order, freqs, btype=btype)
    return sig.lfilter(b, a, x)

def find_butt_bandpass_order(band, fs):
    """
    Claculate the order of Butterworth bandpass filter using minimization of metric between ideal and real frequency response.

    Parameters
    ----------
    band : array_like
        Pair of frequencies. Bounds of bandpass (Hz)
    fs : float
        Sampling rate (Hz)

    Returns
    -------
    order : integer
        Order of filter

    """
    # TODO: Write a set of functions or stubs, all about order
    N = round(60 * 120 * fs)
    unit_pulse = np.zeros(N)
    unit_pulse[1] = 1
    n1 = round((band[0] * N//2) / (fs/2))
    n2 = round((band[1] * N//2) / (fs/2))
    ideal_fr = np.zeros(N)
    f = np.fft.fftfreq(N, 1/fs)
    ind = (f>= band[0])&(f<= band[1])
    ideal_fr[ind] = 1
    ideal_fr = ideal_fr[:N//2]
    prev_metric = np.inf
    for order in range(3, 21):
        impulse_response = butter_filter(unit_pulse, fs, band, order, btype="band")
        fr = abs(fftpack.fft(impulse_response))[:N//2]
        metric = np.sum((fr - ideal_fr)**2)**0.5
        best_order = order
        if (np.isnan(metric)) or (metric >= prev_metric):
            best_order -= 1
            break
        prev_metric = metric
    return best_order-1 # TODO: Think about it

def haar_one_step(x, t, denominator=2):
    """ 
    One cascade of Haar transform.

    Parameters
    ----------
    x : array_like
        Signal values
    t : array_like
        Time values
    denominator : integer
        Denominator used in Haar transform (default is 2)

    Returns
    -------
    x_s : np.array
        Scaled signal values
    x_d : np.array
        Details of x
    t_new : np.array
        Decimated time values

    """
    # TODO: use t or fs in arguments
    x_s = []
    x_d = []
    t_new = []
    for x_left, x_right in zip(x[::2], x[1::2]):
        x_s.append((x_left+x_right)/denominator)
        x_d.append((x_left-x_right)/denominator)
    t_new = t[1::2]
    return np.array(x_s), np.array(x_d), np.array(t_new)

def haar_scaling(x, t, steps_number):
    """ 
    Scaling with Haar transform.

    Parameters
    ----------
    x : array_like
        Signal values
    t : array_like
        Time values
    steps_number : integer
        Number of cascades

    Returns
    -------
    x_s : np.array
        Scaled signal values
    t_new : np.array
        Decimated time values

    """
    # TODO: validation of steps_number
    x_s = x.copy()
    t_new = t.copy()
    for i in range(steps_number):
        x_s, x_d, t_new = haar_one_step(x_s, t_new, denominator=2)
    return x_s, t_new

def trend_smooth(x, fs=1, t=[], cut_off=0.5):
    """
    Calculate trend of signal using smoothing filter.

    Parameters
    ----------
    x : array_like
        Signal values
    t : array_like
        Time values
    cut_off : float
        The frequencies lower than this are trend's frequencies

    Returns
    -------
    trend : np.array
        Trend values
    t_new : np.array
        Time values
    
    """
    if len(t) == 0:
        t = np.linspace(0, (len(x)-1)*fs, len(x))
    else:
        fs = 1.0/(t[1]-t[0])
    win_len = int(fs/2/cut_off)
    if win_len >= len(x):
        return None
    win = np.hamming(win_len)
    trend = np.convolve(x, win, mode="valid") / np.sum(win)
    return trend, t[win_len-1:].copy()



# Copyright (C) 2017-2018 Aleksandr Popov

# This program is free software: you can redistribute it and/or modify
# it under the terms of the Lesser GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# Lesser GNU General Public License for more details.

# You should have received a copy of the Lesser GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# TODO: rename this module

import numpy as np

def freq_by_extremums(x, fs):
    """
    Calculate frequency of oscillating signal by extremums.

    Parameters
    ----------
    x : array_like
        Values of input signals
    fs : float
        Sampling frequency (Hz)

    Returns
    -------
    freq : float
        Frequency

    """
    T = len(x)/fs
    n_max = 0
    n_min = 0
    for x_p, x_c, x_n in zip(x[:-2], x[1:-1], x[2:]):
        if (x_p < x_c) and (x_c >= x_n):
            n_max += 1
        if (x_p > x_c) and (x_c <= x_n):
            n_min += 1
    n = (n_max + n_min) / 2
    return n / T

def freq_by_zeros(x, fs):
    """
    Calculate frequency of oscillating signal by zeros. Signal must be detrended before.

    # TODO: par and ret
    """
    T = len(x)/fs
    n = 0
    for x_p, x_c in zip(x[:-1], x[1:]):
        if x_p * x_c < 0:
            n += 1
    return n / 2 / T

def freqs_stupid(x, fs, window_width = 1024, window_step = 512):
    """
    Calculate an array of frequencies of oscillating signal using window

    Parameters
    ----------
    x : array_like
        Values of input signals
    fs : float
        Sampling frequency (Hz)
    window_width : integer
        Width of window (Samples)
    window_step : integer
        Distance between centers of nearby windows (Samples)

    Returns
    -------
    freqs : np.ndarray
        Frequency values
    t_new : time values

    """
    freqs = []
    t_new = []
    start = 0
    stop = window_width
    if stop > len(x):
        return freqs, t_new
    while True:
        f = freq_by_zeros(x[start : stop], fs)
        freqs.append(f)
        t_new.append((stop-1)/fs)
        start += window_step
        stop += window_step
        if stop > len(x):
            break
    return np.array(freqs), np.array(t_new)

def linint(x, t, t_new, cut_nans=True):
    """
    Find values of x in t_new points.

    Parameters
    ----------
    x : np.ndarray
        Signal values.
    t : np.ndarray
        Time values.
    t_new : np.ndarray
        New time values.
    cat_nans : boolean
        If True, the nan values at the begin and at the end of produced array will removed. 
        Such values will be appear if t_new is wider than t.
 
    Returns
    -------
    x_new : np.ndarray
        New signal values.
 
    """
    x_new = np.zeros(len(t_new)) * np.nan
    for x_p, t_p, x_c, t_c in zip(x[:-1], t[:-1], x[1:], t[1:]):
        k = (x_c - x_p) / (t_c - t_p)
        b = x_p - k*t_p
        ind = (t_new>=t_p)&(t_new<=t_c) 
        x_new[ind] = k*t_new[ind] + b # TODO: rewrite it not using this ind, try to do it real-time
    return x_new

def wave_lens(x, t):
    """
    Calculate wave lengths of signal by space between zeros.

    Parameters
    ----------
    x : np.ndarray
        Signal values.
    t : np.ndarray
        Time values.

    Returns
    -------
    lens : np.ndarray
        Wave lengths.
    t_lent : np.ndarray
        Time values.
    
    """
    tms = []
    for t_c, x_p, x_c in zip(t[1:], x[:-1], x[1:]):
        if x_p * x_c < 0:
            tms.append(t_c)
    # TODO: replace diff by something like real-time, do all in one cycle
    lens = np.diff(tms) * 2
    t_lens = np.array(tms[1:])
    return lens, t_lens

def freqs_by_wave_len(x, t, cut_nans=True):
    """
    Calculate frequencies using lenghs of waves and linear interpolation.

    Parameters
    ----------
    x : np.ndarray
        Signal values.
    t : np.ndarray
        Time values.
    cut_nans : boolean
        If True, the nan values at the ends of the of the produced array will removed. 

    Returns
    -------
    freqs : np.ndarray
        Freqs values
    
    """
    wl, t_wl = wave_lens(x, t)
    freqs = 1/linint(wl, t_wl, t)
    if cut_nans:
        freqs_cut = []
        t_cut = []
        for (f, tt) in zip(freqs, t): # TODO: name tt
            if f==f:
                freqs_cut.append(f)
                t_cut.append(tt)
        return np.array(freqs_cut), t_cut
    return freqs, t




# Copyright (C) 2017-2018 Aleksandr Popov

# This program is free software: you can redistribute it and/or modify
# it under the terms of the Lesser GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# Lesser GNU General Public License for more details.

# You should have received a copy of the Lesser GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

""" Helpers. """

import importlib

def is_iterable(x):
    """ Check if x is iterable. """
    try:
        _ = [e for e in x]
    except TypeError:
        return False
    return True

def import_entity(name):
    """ Import class by name. """
    parts = name.split('.')
    module_name = '.'.join(parts[:-1])
    entity_name = parts[-1]
    if len(module_name) == 0:
        module_name = '__main__'
    module = importlib.import_module(module_name)
    entity = getattr(module, entity_name)
    return entity



# Copyright (C) 2017-2018 Aleksandr Popov

# This program is free software: you can redistribute it and/or modify
# it under the terms of the Lesser GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# Lesser GNU General Public License for more details.

# You should have received a copy of the Lesser GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

""" 
This module implements the Node and Plan classes. Node can be
understood as the workplace for worker. Node can have inputs that are
also nodes. Plan is the system of linked nodes.
"""

from dsplab.activity import Work
from dsplab.helpers import *

class Node:
    """ The node. Node can be understood as the workplace for
    worker. Node can have inputs that are also nodes. """
    def __init__(self, work=None, inputs=[]):
        """ Initialization. """
        self._work = work
        self._res = None
        self.set_inputs(inputs)

    def get_work(self):
        return self._work
    def set_work(self, work):
        self._work = work
    work = property(get_work, set_work)

    def get_inputs(self):
        """ Return inputs. """
        return self._inputs
    def set_inputs(self, inputs):
        """ Set inputs. """
        self._inputs = inputs
    inputs = property(get_inputs, set_inputs)

    def is_output_ready(self) -> bool:
        """ Check if the calculation of data in the node is finished. """
        ans = self._res is not None
        return ans

    def is_inputs_ready(self) -> bool:
        """ Check if data in all inputs is ready. """
        for inpt in self._inputs:
            if not inpt.is_output_ready():
                return False
        return True

    def result(self):
        """ Return the calculated data. """
        return self._res

    def __call__(self, x=None):
        """ Run node. """
        if x is not None:
            y = self.work(x)
            self._res = y
            return
        
        self._res = None
        if len(self._inputs) == 1:
            x = self._inputs[0].result()
        else:
            x = [inpt.result() for inpt in self._inputs]
        y = self.work(x)
        self._res = y
    
class Plan:
    """ The plan. Plan is the system of linked nodes. """
    def __init__(self):
        """ Initialization. """
        super().__init__()
        self._nodes = []
        self._first_nodes = []
        self._last_nodes = []

    def _detect_terminals(self):
        """ Detect first and last nodes. """
        self._first_nodes = []
        all_inputs = []
        for node in self._nodes:
            if len(node.inputs) == 0:
                self._first_nodes.append(node)
            for inpt in node.inputs:
                if inpt not in all_inputs:
                    all_inputs.append(inpt)

        self._last_nodes = []
        for node in self._nodes:
            if node not in all_inputs:
                self._last_nodes.append(node)

    def add_node(self, node, inputs=[]):
        """ Add node to plan. """
        self._nodes.append(node)
        if len(inputs) > 0:
            node.inputs = inputs

    def __call__(self, xs):
        """ Run plan. """
        self._detect_terminals()
        for [node, x] in zip(self._first_nodes, xs):
            node(x)
        
        while True:
            finished = True
            for node in self._nodes:
                if not node.is_output_ready() and node.is_inputs_ready():
                    finished = False
                    node()
            if finished:
                break
            
        ys = [last_node.result() for last_node in self._last_nodes]
        return ys

def setup_plan(plan: Plan, nodes_settings) -> bool:
    """ Setup plan using dict settings. """
    nodes = {}
    
    for node_settings in nodes_settings:
        node_id = node_settings['id']
        nodes[node_id] = Node()
        
        work_settings = node_settings['work']
        if 'descr' in work_settings.keys():
            work_descr = work_settings['descr']
        else:
            work_descr = ""
        
        worker_settings = work_settings['worker']
        worker_class = worker_settings['class']
        
        if 'params' in worker_settings.keys():
            worker_params = worker_settings['params']
            worker = import_entity(worker_class)(**worker_params)
        else:
            worker = import_entity(worker_class)

        work = Work(work_descr, worker)
        nodes[node_id].work = work
        
        if 'inputs' in node_settings.keys():
            input_ids = node_settings['inputs']
            plan.add_node(nodes[node_id], inputs=[nodes[v] for v in input_ids])
        else:
            plan.add_node(nodes[node_id])
        
    return True



# dsplab - Signal processing tools developed by dsplab (http://dsplab.narfu.ru) 

# Copyright (C) 2017-2018 Aleksandr Popov

# This program is free software: you can redistribute it and/or modify
# it under the terms of the Lesser GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# Lesser GNU General Public License for more details.

# You should have received a copy of the Lesser GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from setuptools import setup
import os

from dsplab.glob import *

def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

setup(
    name=PACKAGE,
    version=VERSION,
    description="Some tools for digital signal processing",
    author="Aleksandr Popov",
    author_email="aleneus@gmail.com",
    license = "GPLv3",
    keywords = "digital signal processing",
    url = "https://bitbucket.org/aleneus/dsplab",
    long_description=read('README'),
    packages=['dsplab'],
    install_requires = [
        'numpy',
        'scipy',
        'nose',
    ],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Environment :: Console",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: POSIX :: Linux",
        "Operating System :: Microsoft :: Windows",
        "Programming Language :: Python :: 3",
        "Topic :: Scientific/Engineering",
    ],
)



# Copyright (C) 2017-2018 Aleksandr Popov

# This program is free software: you can redistribute it and/or modify
# it under the terms of the Lesser GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# Lesser GNU General Public License for more details.

# You should have received a copy of the Lesser GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np
import dsplab.filtration as flt

# TODO: [3] add tests
# TODO: [1] allow use t or fs
# TODO: [1] rename f_central to fc (f carrier)
def iq_demod(x, t, f_central, a, b):
    """ Return instantaneous frequency of modulated signal using IQ processign. 

    Parameters
    ----------
    x : array_like
        Signal values.
    t : array_like
        Time values.
    f_central : float
        Carrier frequency.
    a : array_like
        a values of filter.
    b : array_like
        b values of filter.

    Returns
    -------
    freq : np.ndarray of floats
        Instantaneous frequency values.
    t_freq : np.ndarray
        Time values.

    """
    fs = 1/(t[1] - t[0])
    yI = x * np.cos(2*np.pi*f_central*t)
    yQ = x * np.sin(2*np.pi*f_central*t)
    # TODO: [3] next works only for FIR now, some kind of stub
    yI_ = np.convolve(yI, b, mode="same") # TODO: [3] use lfilter and a
    yQ_ = np.convolve(yQ, b, mode="same") # TODO: [3] use lfilter and a
    analytic = yI_ + 1j*yQ_
    phase = -np.unwrap(np.angle(analytic))
    freq = np.diff(phase) / (2*np.pi) * fs + f_central
    return freq, t[:-1]



# Copyright (C) 2017-2018 Aleksandr Popov

# This program is free software: you can redistribute it and/or modify
# it under the terms of the Lesser GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# Lesser GNU General Public License for more details.

# You should have received a copy of the Lesser GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

""" This module implements the base classes for offline and online data processing tools. """

from collections import deque, OrderedDict
import json
import numpy as np

class Activity:
    """ Any activity. Something that may be called and can provide the
    information about itself. """
    def __init__(self):
        """ Initialization. """
        self._info = {}
        self._info['class'] = self.__class__.__name__
    
    def info(self, as_string=False):
        """ Return the information about activity. 
        
        Parameters
        ----------
        as_string: bool
            Method returns JSON-string if True and dict overwise.

        Returns
        -------
        : str or dict
            Information about activity.

        """
        if as_string:
            return json.dumps(self._info, sort_keys=True, indent=4, separators=(',', ': '))
        else:
            return self._info
    
    def __call__(self):
        """ Act. """
        raise NotImplementedError

class OnlineFilter(Activity):
    """ Base class for online filter. """
    def  __init__(self, ntaps=None, smooth_ntaps=None, fill_with=0, step=1):
        """ 
        Initialization.

        Parameters
        ----------
        ntaps : int
            Length of internal queue using for accumulation of input
            samples. Default is None.
        smooth_ntaps : int
            Length of queue using for smoothing output values. Default
            id None.
        fill_with : float
            Initial value of every element of queues.
        step : int
            Step. Must be positive.

        """
        super().__init__()
        # TODO: add arguments to self._info
        self.add_sample_func = None
        if   (ntaps == None) and (smooth_ntaps == None):
            self.add_sample_func = self.__add_sample_simple
        elif (ntaps != None) and (smooth_ntaps == None):
            self.add_sample_func = self.__add_sample_only_queue
        elif (ntaps == None) and (smooth_ntaps != None):
            self.add_sample_func = self.__add_sample_only_smooth
        else:
            self.add_sample_func = self.__add_sample_full

        self.queue = None
        self.smooth_queue = None
        if ntaps!=None:
            self.queue = deque([fill_with]*ntaps, maxlen=ntaps)
        if smooth_ntaps!=None:
            self.smooth_queue = deque([fill_with]*smooth_ntaps, maxlen=smooth_ntaps)
            wind = np.hamming(smooth_ntaps)
            self.wind = wind / sum(wind)

        self.step = step
        self.steps = 0

        self.ntaps = ntaps
        self.smooth_ntaps = smooth_ntaps

    def add_sample(self, x):
        """
        Add input sample to filter and return output value.

        Parameters
        ----------
        x : float
            Input value.
        
        Returns
        -------
        : float
            Output value.

        """
        # TODO: depracate, use __call__
        return self.add_sample_func(x)

    def __call__(self, x):
        return self.add_sample(x)

    def __add_sample_simple(self, x):
        """ Add sample without using queues. """
        self.steps += 1
        if self.steps == self.step:
            self.steps = 0
            return self.proc_sample(x)
        return None

    def __add_sample_only_queue(self, x):
        """ Add sample with no smoothing. """
        self.steps += 1
        self.queue.append(x)
        if self.steps == self.step:
            self.steps = 0
            return self.proc_queue()
        return None

    def __add_sample_only_smooth(self, x):
        """ Add sample with not internal queue but with smoothed ouput. """
        self.steps += 1
        if self.steps == self.step:
            self.steps = 0
            self.smooth_queue.append(self.proc_sample(x))
            resm = np.dot(np.array(self.smooth_queue), self.wind)
            return resm
        return None

    def __add_sample_full(self, x):
        """ Add sample with internal queue and smoothing of ouput values. """
        self.steps += 1
        self.queue.append(x)
        if self.steps == self.step:
            self.steps = 0
            self.smooth_queue.append(self.proc_queue())
            resm = np.dot(np.array(self.smooth_queue), self.wind)
            return resm
        return None

    def proc_queue(self):
        """ 
        Process queue.

        Returns
        -------
        : float
            Ouput value.

        """
        raise NotImplementedError

    def proc_sample(self, x):
        """
        Process sample.

        Parameters
        ----------
        x : float
            Input value.
        
        Returns
        -------
        : float
            Output value.

        """
        raise NotImplementedError
    
class OnlineLogic(OnlineFilter):
    """ Base class for logical connectors of outputs of several
    detectors or other connectors. """
    # TODO: test info
    def __init__(self, inputs=[]):
        """ 
        Initialization.

        Parameters
        ----------
        inputs : list
            Input filters.

        """
        super().__init__()
        self._info['descr'] = 'Logical operation'
        self._info['inputs'] = []
        self.inputs = []
        for inpt in inputs:
            self.add_input(inpt)

    def add_input(self, inpt):
        """ Add input of other connector.

        Parameters
        ----------
        inpt : Object
            Input.

        """
        self.inputs.append(inpt)
        self._info['inputs'].append(inpt.info())

    def add_sample(self, xs):
        raise NotImplementedError

class And(OnlineLogic):
    """ And connector. """
    def add_sample(self, xs):
        result = 1
        for (inpt, x) in zip(self.inputs, xs):
            result *= inpt.add_sample(x)
        return result

class Or(OnlineLogic):
    """ Or connector. """
    def add_sample(self, xs):
        res = 0
        for (inpt, x) in zip(self.inputs, xs):
            res += inpt.add_sample(x) * (1 - res)
        return res

class Work(Activity):
    """ Work is activity which has some worker. Different workers can
    be used for doing the same work. """
    def __init__(self, descr="", worker=None):
        """ Initialization. """
        super().__init__()
        self.set_descr(descr)
        self.set_worker(worker)

    def set_descr(self, descr):
        """ Set description of work. """
        self.descr = descr
        self._info['descr'] = descr

    def set_worker(self, worker):
        """ Set worker for doing work. """
        self.worker = worker
        self._info['worker'] = None
        try:
            self._info['worker'] = worker.info()
        except:
            pass

    def __call__(self, *args, **kwargs):
        """ Do work. """
        y = self.worker(*args, *kwargs)
        return y

class Strategy(Activity):
    """ Deprecated. """
    def __init__(self, name="", info=""):
        """ Initialization. """
        print("Deprecated: dsplab.activity.Strategy")
        self.name = name
        self.workers = OrderedDict()

    def set_worker(self, work, worker):
        """
        Add worker.
        
        Parameters
        ----------
        work : str
            Name of work.
        worker : object
            Worker object.
        
        """
        self.workers[work] = worker

    def __call__(self):
        raise NotImplementedError

class LinearStrategy(Strategy):
    """ Linear strategy. Works called one by one, from first setted
    work to the last one. """
    def __call__(self, x):
        y = x
        print(self.workers)
        for work in self.workers:
            y = self.workers[work](y)
        return y

if __name__ == "__main__":
    or_connector = Or()
    print(or_connector.info(as_string=True))
